import unittest
import logging
import sys
import time
from unittest.mock import patch, Mock

import graphsignal
from graphsignal.samplers.random_sampling import RandomSampler

logger = logging.getLogger('graphsignal')


class RandomSamplerTest(unittest.TestCase):
    def setUp(self):
        graphsignal.configure(
            api_key='k1',
            deployment='d1',
            upload_on_shutdown=False,
            debug_mode=True)

    def tearDown(self):
        graphsignal.shutdown()

    def test_sample(self):
        sampler = RandomSampler()

        for _ in range(1000):
            self.assertTrue(sampler.sample())

        self.assertFalse(sampler.sample())

        sampler._last_reset_ts = time.time() - 61

        self.assertTrue(sampler.sample())



import unittest
import logging
import sys
import os
import json
import time
from unittest.mock import patch, Mock
import chromadb

import graphsignal
from graphsignal.proto import signals_pb2
from graphsignal.uploader import Uploader
from graphsignal.spans import DEFAULT_OPTIONS
from graphsignal.recorders.chroma_recorder import ChromaRecorder
from graphsignal.proto_utils import find_tag, find_param, find_data_count, find_data_sample

logger = logging.getLogger('graphsignal')


class ChromaRecorderTest(unittest.IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        if len(logger.handlers) == 0:
            logger.addHandler(logging.StreamHandler(sys.stdout))
        graphsignal.configure(
            api_key='k1',
            deployment='d1',
            upload_on_shutdown=False,
            debug_mode=True)

    async def asyncTearDown(self):
        graphsignal.shutdown()

    async def test_record(self):
        recorder = ChromaRecorder()
        recorder.setup()
        proto = signals_pb2.Span()
        context = {}
        recorder.on_span_start(proto, context, DEFAULT_OPTIONS)
        recorder.on_span_stop(proto, context, DEFAULT_OPTIONS)
        recorder.on_span_read(proto, context, DEFAULT_OPTIONS)

        self.assertEqual(proto.libraries[0].name, 'Chroma')

    @patch.object(Uploader, 'upload_span')
    async def test_trace_add(self, mocked_upload_span):
        client = chromadb.Client()

        collection = client.get_or_create_collection('test_col')

        collection.upsert(
            embeddings=[[1.5, 2.9, 3.4], [9.8, 2.3, 2.9]],
            documents=["doc1", "doc2"],
            metadatas=[{"source": "my_source"}, {"source": "my_source"}],
            ids=["id1", "id2"],
            increment_index=True
        )

        collection.get(
            ids=["id1", "id2"]
        )

        collection.query(
            query_embeddings=[[1.1, 2.3, 3.2], [5.1, 4.3, 2.2]],
            n_results=2
        )

        collection.delete()

        upsert_proto = mocked_upload_span.call_args_list[0][0][0]
        get_proto = mocked_upload_span.call_args_list[1][0][0]
        query_proto = mocked_upload_span.call_args_list[2][0][0]
        delete_proto = mocked_upload_span.call_args_list[3][0][0]

        # upsert
        self.assertEqual(upsert_proto.libraries[0].name, 'Chroma')
        self.assertEqual(find_tag(upsert_proto, 'component'), 'Memory')
        self.assertEqual(find_tag(upsert_proto, 'operation'), 'chroma.collection.upsert')
        self.assertEqual(find_tag(upsert_proto, 'index'), 'test_col')
        self.assertEqual(find_param(upsert_proto, 'collection'), 'test_col')
        self.assertEqual(find_data_count(upsert_proto, 'embeddings', 'element_count'), 6.0)

        # get
        self.assertEqual(get_proto.libraries[0].name, 'Chroma')
        self.assertEqual(find_tag(get_proto, 'component'), 'Memory')
        self.assertEqual(find_tag(get_proto, 'operation'), 'chroma.collection.get')
        self.assertEqual(find_tag(get_proto, 'index'), 'test_col')
        self.assertEqual(find_param(get_proto, 'collection'), 'test_col')
        self.assertEqual(find_data_count(get_proto, 'documents', 'element_count'), 2.0)
        self.assertIsNotNone(find_data_sample(get_proto, 'documents'))

        # query
        self.assertEqual(query_proto.libraries[0].name, 'Chroma')
        self.assertEqual(find_tag(query_proto, 'component'), 'Memory')
        self.assertEqual(find_tag(query_proto, 'operation'), 'chroma.collection.query')
        self.assertEqual(find_tag(query_proto, 'index'), 'test_col')
        self.assertEqual(find_param(query_proto, 'collection'), 'test_col')
        self.assertEqual(find_data_count(query_proto, 'query_embeddings', 'element_count'), 6.0)
        self.assertEqual(find_data_count(get_proto, 'documents', 'element_count'), 2.0)
        self.assertIsNotNone(find_data_sample(get_proto, 'documents'))

        # delete
        self.assertEqual(delete_proto.libraries[0].name, 'Chroma')
        self.assertEqual(find_tag(delete_proto, 'component'), 'Memory')
        self.assertEqual(find_tag(delete_proto, 'operation'), 'chroma.collection.delete')
        self.assertEqual(find_tag(delete_proto, 'index'), 'test_col')
        self.assertEqual(find_param(delete_proto, 'collection'), 'test_col')



import unittest
import logging
import sys
import os
import json
import time
from unittest.mock import patch, Mock
from google.protobuf.json_format import MessageToJson
import pprint
import types
import openai

import graphsignal
from graphsignal.proto import signals_pb2
from graphsignal.uploader import Uploader
from graphsignal.spans import DEFAULT_OPTIONS
from graphsignal.recorders.openai_recorder import OpenAIRecorder
from graphsignal.proto_utils import find_tag, find_param, find_data_count, find_data_sample
from graphsignal.proto_utils import parse_semver, compare_semver
from graphsignal.proto import signals_pb2

logger = logging.getLogger('graphsignal')

openai_lib = signals_pb2.LibraryInfo()
parse_semver(openai_lib.version, openai.version.VERSION)

if compare_semver(openai_lib.version, (1, 0, 0)) < 0:
    class OpenAIRecorderTest(unittest.IsolatedAsyncioTestCase):
        async def asyncSetUp(self):
            if len(logger.handlers) == 0:
                logger.addHandler(logging.StreamHandler(sys.stdout))
            graphsignal.configure(
                api_key='k1',
                deployment='d1',
                upload_on_shutdown=False,
                debug_mode=True)

        async def asyncTearDown(self):
            graphsignal.shutdown()

        async def test_record(self):
            recorder = OpenAIRecorder()
            recorder.setup()
            proto = signals_pb2.Span()
            context = {}
            recorder.on_span_start(proto, context, DEFAULT_OPTIONS)
            recorder.on_span_stop(proto, context, DEFAULT_OPTIONS)
            recorder.on_span_read(proto, context, DEFAULT_OPTIONS)
            self.assertEqual(proto.libraries[0].name, 'OpenAI Python Library')

        @patch.object(Uploader, 'upload_span')
        @patch.object(openai.Completion, 'create')
        async def test_completion_create(self, mocked_create, mocked_upload_span):
            # mocking overrides autoinstrumentation, reinstrument
            recorder = OpenAIRecorder()
            recorder.setup()

            mocked_create.return_value = {
                "choices": [
                    {
                        "finish_reason": "stop",
                        "index": 0,
                        "logprobs": None,
                        "text": "\n\ntest completion 1"
                    },
                    {
                        "finish_reason": "length",
                        "index": 1,
                        "logprobs": None,
                        "text": "\n\ntest completion 222"
                    }
                ],
                "created": 1675070122,
                "id": "cmpl-id",
                "model": "text-davinci-003",
                "object": "text_completion",
                "usage": {
                    "completion_tokens": 96,
                    "prompt_tokens": 6,
                    "total_tokens": 102
                }
            }

            response = openai.Completion.create(
                model="text-davinci-003", 
                prompt=['test prompt 1', 'test prompt 2'],
                temperature=0.1,
                top_p=1,
                max_tokens=1024,
                frequency_penalty=0,
                presence_penalty=0)

            recorder.shutdown()

            proto = mocked_upload_span.call_args[0][0]

            self.assertEqual(find_tag(proto, 'component'), 'LLM')
            self.assertEqual(find_tag(proto, 'operation'), 'openai.Completion.create')
            self.assertEqual(find_tag(proto, 'endpoint'), 'https://api.openai.com/v1/completions')
            self.assertEqual(find_tag(proto, 'model'), 'text-davinci-003')

            self.assertEqual(find_param(proto, 'model'), 'text-davinci-003')
            self.assertEqual(find_param(proto, 'max_tokens'), '1024')
            self.assertEqual(find_param(proto, 'temperature'), '0.1')
            self.assertEqual(find_param(proto, 'top_p'), '1')
            self.assertEqual(find_param(proto, 'presence_penalty'), '0')
            self.assertEqual(find_param(proto, 'frequency_penalty'), '0')

            self.assertEqual(find_data_count(proto, 'prompt', 'token_count'), 6.0)
            self.assertEqual(find_data_count(proto, 'completion', 'token_count'), 96.0)
            self.assertEqual(find_data_count(proto, 'completion', 'finish_reason_stop'), 1.0)
            self.assertEqual(find_data_count(proto, 'completion', 'finish_reason_length'), 1.0)

            self.assertIsNotNone(find_data_sample(proto, 'prompt'))
            self.assertIsNotNone(find_data_sample(proto, 'completion'))

        @patch.object(Uploader, 'upload_span')
        @patch.object(openai.Completion, 'create')
        async def test_completion_create_stream(self, mocked_create, mocked_upload_span):
            # mocking overrides autoinstrumentation, reinstrument
            recorder = OpenAIRecorder()
            recorder.setup()

            test_ret = [
                {
                    "choices": [
                        {
                        "finish_reason": None,
                        "index": 1,
                        "logprobs": None,
                        "text": "abc"
                        }
                    ],
                    "created": 1676896808,
                    "id": "cmpl-6lzkernKqOvF4ewZGhHlZ63HZJcbc",
                    "model": "text-davinci-003",
                    "object": "text_completion"
                },
                {
                    "choices": [
                        {
                        "finish_reason": "stop",
                        "index": 1,
                        "logprobs": None,
                        "text": "\n"
                        }
                    ],
                    "created": 1676896808,
                    "id": "cmpl-6lzkernKqOvF4ewZGhHlZ63HZJcbc",
                    "model": "text-davinci-003",
                    "object": "text_completion"
                }
            ]
            def test_ret_gen():
                for item in test_ret:
                    yield item
            mocked_create.return_value = test_ret_gen()

            response = openai.Completion.create(
                model="text-davinci-003", 
                prompt=['count 1 to 3', 'generate 2 random letters'],
                temperature=0.1,
                top_p=1,
                max_tokens=1024,
                frequency_penalty=0,
                presence_penalty=0,
                stream=True)

            for r in response:
                pass

            recorder.shutdown()

            proto = mocked_upload_span.call_args[0][0]

            self.assertEqual(find_data_count(proto, 'prompt', 'token_count'), 9.0)
            self.assertEqual(find_data_count(proto, 'completion', 'finish_reason_stop'), 1.0)
            self.assertEqual(find_data_count(proto, 'completion', 'token_count'), 2.0)

            self.assertIsNotNone(find_data_sample(proto, 'prompt'))
            self.assertIsNotNone(find_data_sample(proto, 'completion'))

        @patch.object(Uploader, 'upload_span')
        @patch.object(openai.Completion, 'acreate')
        async def test_completion_acreate(self, mocked_create, mocked_upload_span):
            # mocking overrides autoinstrumentation, reinstrument
            recorder = OpenAIRecorder()
            recorder.setup()

            mocked_create.return_value = {
                "choices": [
                    {
                        "finish_reason": "stop",
                        "index": 0,
                        "logprobs": None,
                        "text": "\n\ntest completion 1"
                    },
                    {
                        "finish_reason": "stop",
                        "index": 1,
                        "logprobs": None,
                        "text": "\n\ntest completion 222"
                    }
                ],
                "created": 1675070122,
                "id": "cmpl-id",
                "model": "text-davinci-003",
                "object": "text_completion",
                "usage": {
                    "completion_tokens": 96,
                    "prompt_tokens": 6,
                    "total_tokens": 102
                }
            }

            response = await openai.Completion.acreate(
                model="text-davinci-003", 
                prompt=['test prompt 1', 'test prompt 2'],
                temperature=0.1,
                top_p=1,
                max_tokens=1024,
                frequency_penalty=0,
                presence_penalty=0)

            recorder.shutdown()

            proto = mocked_upload_span.call_args[0][0]

            self.assertEqual(find_tag(proto, 'component'), 'LLM')
            self.assertEqual(find_tag(proto, 'operation'), 'openai.Completion.acreate')
            self.assertEqual(find_tag(proto, 'endpoint'), 'https://api.openai.com/v1/completions')
            self.assertEqual(find_tag(proto, 'model'), 'text-davinci-003')

            self.assertEqual(find_param(proto, 'model'), 'text-davinci-003')
            self.assertEqual(find_param(proto, 'max_tokens'), '1024')
            self.assertEqual(find_param(proto, 'temperature'), '0.1')
            self.assertEqual(find_param(proto, 'top_p'), '1')
            self.assertEqual(find_param(proto, 'presence_penalty'), '0')
            self.assertEqual(find_param(proto, 'frequency_penalty'), '0')

            self.assertEqual(find_data_count(proto, 'prompt', 'token_count'), 6.0)
            self.assertEqual(find_data_count(proto, 'completion', 'token_count'), 96.0)

            self.assertIsNotNone(find_data_sample(proto, 'prompt'))
            self.assertIsNotNone(find_data_sample(proto, 'completion'))

        @patch.object(Uploader, 'upload_span')
        @patch.object(openai.ChatCompletion, 'create')
        async def test_chat_completion_create(self, mocked_create, mocked_upload_span):
            # mocking overrides autoinstrumentation, reinstrument
            recorder = OpenAIRecorder()
            recorder.setup()

            mocked_create.return_value = {
                "choices": [
                    {
                        "finish_reason": "stop",
                        "index": 0,
                        "logprobs": None,
                        "message": {
                            "content": "\n\ntest completion 1"
                        }
                    },
                    {
                        "finish_reason": "length",
                        "index": 1,
                        "logprobs": None,
                        "message": {
                            "content": "\n\ntest completion 222"
                        }
                    }
                ],
                "created": 1675070122,
                "id": "cmpl-id",
                "model": "gpt-3.5-turbo",
                "object": "text_completion",
                "usage": {
                    "completion_tokens": 96,
                    "prompt_tokens": 6,
                    "total_tokens": 102
                }
            }

            response = openai.ChatCompletion.create(
                model="gpt-3.5-turbo", 
                messages=[
                    {'role': 'system', 'content': 'test prompt 1'}, 
                    {'role': 'system', 'content': 'test prompt 2'}],
                temperature=0.1,
                top_p=1,
                max_tokens=1024,
                frequency_penalty=0,
                presence_penalty=0)

            recorder.shutdown()

            proto = mocked_upload_span.call_args[0][0]

            self.assertEqual(find_tag(proto, 'component'), 'LLM')
            self.assertEqual(find_tag(proto, 'operation'), 'openai.ChatCompletion.create')
            self.assertEqual(find_tag(proto, 'endpoint'), 'https://api.openai.com/v1/chat/completions')
            self.assertEqual(find_tag(proto, 'model'), 'gpt-3.5-turbo')

            self.assertEqual(find_param(proto, 'model'), 'gpt-3.5-turbo')
            self.assertEqual(find_param(proto, 'max_tokens'), '1024')
            self.assertEqual(find_param(proto, 'temperature'), '0.1')
            self.assertEqual(find_param(proto, 'top_p'), '1')
            self.assertEqual(find_param(proto, 'presence_penalty'), '0')
            self.assertEqual(find_param(proto, 'frequency_penalty'), '0')

            self.assertEqual(find_data_count(proto, 'messages', 'token_count'), 6.0)
            self.assertEqual(find_data_count(proto, 'completion', 'token_count'), 96.0)
            self.assertEqual(find_data_count(proto, 'completion', 'finish_reason_stop'), 1.0)
            self.assertEqual(find_data_count(proto, 'completion', 'finish_reason_length'), 1.0)

            self.assertIsNotNone(find_data_sample(proto, 'messages'))
            self.assertIsNotNone(find_data_sample(proto, 'completion'))


        @patch.object(Uploader, 'upload_span')
        @patch.object(openai.ChatCompletion, 'create')
        async def test_chat_completion_create_function(self, mocked_create, mocked_upload_span):
            # mocking overrides autoinstrumentation, reinstrument
            recorder = OpenAIRecorder()
            recorder.setup()

            mocked_create.return_value = {
                "choices": [
                    {
                        "finish_reason": "function_call",
                        "index": 0,
                        "message": {
                            "content": None,
                            "function_call": {
                                "name": "get_current_weather",
                                "arguments": "{\n  \"location\": \"Boston, MA\"\n}"
                            }
                        },
                    }
                ],
                "created": 1675070122,
                "id": "chatcmpl-id",
                "model": "gpt-3.5-turbo-0613",
                "object": "chat.completion",
                "usage": {
                    "completion_tokens": 78,
                    "prompt_tokens": 18,
                    "total_tokens": 96
                }
            }

            response = openai.ChatCompletion.create(
                model="gpt-3.5-turbo-0613", 
                messages=[{'role': 'system', 'content': 'What\'s the weather like in Boston?'}],
                functions=[
                    {
                        "name": "get_current_weather",
                        "description": "Get the current weather in a given location",
                        "parameters": {
                            "type": "object",
                            "properties": {
                                "location": {
                                    "type": "string",
                                    "description": "The city and state, e.g. San Francisco, CA",
                                },
                                "unit": {"type": "string", "enum": ["celsius", "fahrenheit"]},
                            },
                            "required": ["location"],
                        },
                    }
                ],
                function_call="auto",
                temperature=0.1,
                top_p=1,
                max_tokens=1024,
                frequency_penalty=0,
                presence_penalty=0)

            recorder.shutdown()

            proto = mocked_upload_span.call_args[0][0]

            self.assertEqual(find_tag(proto, 'component'), 'LLM')
            self.assertEqual(find_tag(proto, 'operation'), 'openai.ChatCompletion.create')
            self.assertEqual(find_tag(proto, 'endpoint'), 'https://api.openai.com/v1/chat/completions')
            self.assertEqual(find_tag(proto, 'model'), 'gpt-3.5-turbo-0613')

            self.assertEqual(find_param(proto, 'model'), 'gpt-3.5-turbo-0613')
            self.assertEqual(find_param(proto, 'function_call'), 'auto')
            self.assertEqual(find_param(proto, 'max_tokens'), '1024')
            self.assertEqual(find_param(proto, 'temperature'), '0.1')
            self.assertEqual(find_param(proto, 'top_p'), '1')
            self.assertEqual(find_param(proto, 'presence_penalty'), '0')
            self.assertEqual(find_param(proto, 'frequency_penalty'), '0')

            self.assertEqual(find_data_count(proto, 'messages', 'token_count'), 18.0)
            self.assertEqual(find_data_count(proto, 'completion', 'token_count'), 78.0)
            self.assertEqual(find_data_count(proto, 'completion', 'finish_reason_function_call'), 1.0)

            self.assertIsNotNone(find_data_sample(proto, 'messages'))
            self.assertIsNotNone(find_data_sample(proto, 'functions'))
            self.assertIsNotNone(find_data_sample(proto, 'completion'))

        @patch.object(Uploader, 'upload_span')
        @patch.object(openai.ChatCompletion, 'create')
        async def test_chat_completion_create_stream(self, mocked_create, mocked_upload_span):
            # mocking overrides autoinstrumentation, reinstrument
            recorder = OpenAIRecorder()
            recorder.setup()

            test_ret = [
                {
                    "choices": [
                        {
                            "finish_reason": None,
                            "index": 1,
                            "logprobs": None,
                            "delta": {
                                "content": "abc"
                            }
                        }
                    ],
                    "created": 1676896808,
                    "id": "cmpl-6lzkernKqOvF4ewZGhHlZ63HZJcbc",
                    "model": "gpt-3.5-turbo",
                    "object": "chat.completion.chunk"
                },
                {
                    "choices": [
                        {
                            "finish_reason": "stop",
                            "index": 1,
                            "logprobs": None,
                            "delta": {
                                "content": "\n"
                            }
                        }
                    ],
                    "created": 1676896808,
                    "id": "cmpl-6lzkernKqOvF4ewZGhHlZ63HZJcbc",
                    "model": "gpt-3.5-turbo",
                    "object": "chat.completion.chunk"
                }
            ]
            def test_ret_gen():
                for item in test_ret:
                    yield item
            mocked_create.return_value = test_ret_gen()

            response = openai.ChatCompletion.create(
                model="gpt-4", 
                messages=[
                    {
                        "role": "system",
                        "content": "You are a helpful.",
                    },
                    {
                        "role": "system",
                        "name": "example_user",
                        "content": "New synergies will help drive top-line growth.",
                    },
                    {
                        "role": "user",
                        "name": "example",
                        "content": "This late pivot means.",
                    }
                ],
                temperature=0.1,
                top_p=1,
                max_tokens=1024,
                frequency_penalty=0,
                presence_penalty=0,
                stream=True)

            for r in response:
                pass

            recorder.shutdown()

            proto = mocked_upload_span.call_args[0][0]

            self.assertEqual(find_data_count(proto, 'messages', 'token_count'), 40.0)
            self.assertEqual(find_data_count(proto, 'completion', 'finish_reason_stop'), 1.0)
            self.assertEqual(find_data_count(proto, 'completion', 'token_count'), 2.0)

            self.assertIsNotNone(find_data_sample(proto, 'messages'))
            self.assertIsNotNone(find_data_sample(proto, 'completion'))


        @patch.object(Uploader, 'upload_span')
        @patch.object(openai.ChatCompletion, 'acreate')
        async def test_chat_completion_acreate_stream(self, mocked_acreate, mocked_upload_span):
            # mocking overrides autoinstrumentation, reinstrument
            recorder = OpenAIRecorder()
            recorder.setup()

            test_ret = [
                {
                    "choices": [
                        {
                            "finish_reason": None,
                            "index": 1,
                            "logprobs": None,
                            "delta": {
                                "content": "abc"
                            }
                        }
                    ],
                    "created": 1676896808,
                    "id": "cmpl-6lzkernKqOvF4ewZGhHlZ63HZJcbc",
                    "model": "gpt-3.5-turbo",
                    "object": "chat.completion.chunk"
                },
                {
                    "choices": [
                        {
                            "finish_reason": "stop",
                            "index": 1,
                            "logprobs": None,
                            "delta": {
                                "content": "\n"
                            }
                        }
                    ],
                    "created": 1676896808,
                    "id": "cmpl-6lzkernKqOvF4ewZGhHlZ63HZJcbc",
                    "model": "gpt-3.5-turbo",
                    "object": "chat.completion.chunk"
                }
            ]
            async def test_ret_gen():
                for item in test_ret:
                    yield item
            mocked_acreate.return_value = test_ret_gen()

            response = await openai.ChatCompletion.acreate(
                model="gpt-4", 
                messages=[
                    {
                        "role": "system",
                        "content": "You are a helpful.",
                    },
                    {
                        "role": "system",
                        "name": "example_user",
                        "content": "New synergies will help drive top-line growth.",
                    },
                    {
                        "role": "user",
                        "name": "example",
                        "content": "This late pivot means.",
                    }
                ],
                temperature=0.1,
                top_p=1,
                max_tokens=1024,
                frequency_penalty=0,
                presence_penalty=0,
                stream=True)

            async for r in response:
                pass

            recorder.shutdown()

            proto = mocked_upload_span.call_args[0][0]

            self.assertEqual(find_data_count(proto, 'messages', 'token_count'), 40.0)
            self.assertEqual(find_data_count(proto, 'completion', 'finish_reason_stop'), 1.0)
            self.assertEqual(find_data_count(proto, 'completion', 'token_count'), 2.0)

            self.assertIsNotNone(find_data_sample(proto, 'messages'))
            self.assertIsNotNone(find_data_sample(proto, 'completion'))

        @patch.object(Uploader, 'upload_span')
        @patch.object(openai.Edit, 'create')
        async def test_edits_create(self, mocked_create, mocked_upload_span):
            # mocking overrides autoinstrumentation, reinstrument
            recorder = OpenAIRecorder()
            recorder.setup()

            mocked_create.return_value = {
                "choices": [
                    {
                        "index": 0,
                        "text": "test input 1\ntest instruction 1\n"
                    }
                ],
                "created": 1675077592,
                "object": "edit",
                "usage": {
                    "completion_tokens": 13,
                    "prompt_tokens": 18,
                    "total_tokens": 32
                }
            }

            response = openai.Edit.create(
                model="text-davinci-edit-001", 
                input='test input 1',
                instruction='test instruction 1',
                temperature=0.1,
                top_p=1)

            recorder.shutdown()

            proto = mocked_upload_span.call_args[0][0]

            self.assertEqual(find_tag(proto, 'component'), 'LLM')
            self.assertEqual(find_tag(proto, 'operation'), 'openai.Edit.create')
            self.assertEqual(find_tag(proto, 'endpoint'), 'https://api.openai.com/v1/edits')
            self.assertEqual(find_tag(proto, 'model'), 'text-davinci-edit-001')

            self.assertEqual(find_param(proto, 'model'), 'text-davinci-edit-001')
            self.assertEqual(find_param(proto, 'temperature'), '0.1')
            self.assertEqual(find_param(proto, 'top_p'), '1')

            self.assertEqual(find_data_count(proto, 'instruction', 'token_count'), 18.0)
            self.assertEqual(find_data_count(proto, 'edits', 'token_count'), 13.0)

            self.assertIsNotNone(find_data_sample(proto, 'input'))
            self.assertIsNotNone(find_data_sample(proto, 'instruction'))
            self.assertIsNotNone(find_data_sample(proto, 'edits'))

        @patch.object(Uploader, 'upload_span')
        @patch.object(openai.Embedding, 'create')
        async def test_embedding_create(self, mocked_create, mocked_upload_span):
            # mocking overrides autoinstrumentation, reinstrument
            recorder = OpenAIRecorder()
            recorder.setup()

            mocked_create.return_value = {
                "data": [
                    {
                        "embedding": [
                            0.0007817253354005516,
                            -0.004567846190184355,
                            -0.011692041531205177
                        ],
                        "index": 0,
                        "object": "embedding"
                        },
                        {
                        "embedding": [
                            -0.008782976306974888,
                            -0.0015969815431162715,
                            -0.004394866060465574
                        ],
                        "index": 1,
                        "object": "embedding"
                        }
                    ],
                    "model": "text-embedding-ada-002",
                    "object": "list",
                    "usage": {
                        "prompt_tokens": 8,
                        "total_tokens": 8
                    }
                }

            response = openai.Embedding.create(
                model="text-embedding-ada-002",
                input=['test text 1', 'test text 2'])

            recorder.shutdown()

            proto = mocked_upload_span.call_args[0][0]

            self.assertEqual(find_tag(proto, 'component'), 'LLM')
            self.assertEqual(find_tag(proto, 'operation'), 'openai.Embedding.create')
            self.assertEqual(find_tag(proto, 'endpoint'), 'https://api.openai.com/v1/embeddings')
            self.assertEqual(find_tag(proto, 'model'), 'text-embedding-ada-002')

            self.assertEqual(find_param(proto, 'model'), 'text-embedding-ada-002')

            self.assertEqual(find_data_count(proto, 'input', 'token_count'), 8.0)

            self.assertIsNotNone(find_data_sample(proto, 'input'))
            self.assertIsNotNone(find_data_sample(proto, 'embeddings'))

        @patch.object(Uploader, 'upload_span')
        @patch.object(openai.Image, 'create')
        async def test_image_create(self, mocked_create, mocked_upload_span):
            # mocking overrides autoinstrumentation, reinstrument
            recorder = OpenAIRecorder()
            recorder.setup()

            mocked_create.return_value = {
                "created": 1675079434,
                "data": [
                    {
                        "b64_json": "iVBORw0KGgoAAA"
                    }
                ]
            }

            response = openai.Image.create(
                prompt='test image',
                n=1,
                size='256x256',
                response_format='b64_json')

            recorder.shutdown()

            proto = mocked_upload_span.call_args[0][0]

            self.assertEqual(find_tag(proto, 'component'), 'Model')
            self.assertEqual(find_tag(proto, 'operation'), 'openai.Image.create')
            self.assertEqual(find_tag(proto, 'endpoint'), 'https://api.openai.com/v1/images/generations')

            self.assertEqual(find_param(proto, 'n'), '1')
            self.assertEqual(find_param(proto, 'size'), '256x256')
            self.assertEqual(find_param(proto, 'response_format'), 'b64_json')

            self.assertEqual(find_data_count(proto, 'image', 'element_count'), 1.0)

        @patch.object(Uploader, 'upload_span')
        @patch.object(openai.Audio, 'transcribe')
        async def test_audio_transcription(self, mocked_create, mocked_upload_span):
            # mocking overrides autoinstrumentation, reinstrument
            recorder = OpenAIRecorder()
            recorder.setup()

            mocked_create.return_value = {
                "text": 'some text'
            }

            import tempfile
            with tempfile.NamedTemporaryFile() as tmp_file:
                tmp_file.write('test content'.encode('utf-8'))
                tmp_file.flush()

                response = openai.Audio.transcribe(
                    file=tmp_file,
                    model='whisper-1',
                    response_format='json',
                    prompt='test prompt',
                    temperature=0.1,
                    language='en')

            recorder.shutdown()

            proto = mocked_upload_span.call_args[0][0]

            self.assertEqual(find_tag(proto, 'component'), 'Model')
            self.assertEqual(find_tag(proto, 'operation'), 'openai.Audio.transcribe')
            self.assertEqual(find_tag(proto, 'endpoint'), 'https://api.openai.com/v1/audio/transcriptions')

            self.assertEqual(find_param(proto, 'model'), 'whisper-1')
            self.assertEqual(find_param(proto, 'response_format'), 'json')
            self.assertEqual(find_param(proto, 'temperature'), '0.1')
            self.assertEqual(find_param(proto, 'language'), 'en')

            self.assertEqual(find_data_count(proto, 'file', 'byte_count'), 12.0)

            self.assertIsNotNone(find_data_sample(proto, 'prompt'))


        @patch.object(Uploader, 'upload_span')
        @patch.object(openai.Audio, 'translate')
        async def test_audio_translate(self, mocked_create, mocked_upload_span):
            # mocking overrides autoinstrumentation, reinstrument
            recorder = OpenAIRecorder()
            recorder.setup()

            mocked_create.return_value = {
                "text": 'some text'
            }

            import tempfile
            with tempfile.NamedTemporaryFile() as tmp_file:
                tmp_file.write('test content'.encode('utf-8'))
                tmp_file.flush()

                response = openai.Audio.translate(
                    file=tmp_file,
                    model='whisper-1',
                    response_format='json',
                    prompt='test prompt',
                    temperature=0.1)

            recorder.shutdown()

            proto = mocked_upload_span.call_args[0][0]

            self.assertEqual(find_tag(proto, 'component'), 'Model')
            self.assertEqual(find_tag(proto, 'operation'), 'openai.Audio.translate')
            self.assertEqual(find_tag(proto, 'endpoint'), 'https://api.openai.com/v1/audio/translations')

            self.assertEqual(find_param(proto, 'model'), 'whisper-1')
            self.assertEqual(find_param(proto, 'response_format'), 'json')
            self.assertEqual(find_param(proto, 'temperature'), '0.1')

            self.assertEqual(find_data_count(proto, 'file', 'byte_count'), 12.0)

            self.assertIsNotNone(find_data_sample(proto, 'prompt'))

        @patch.object(Uploader, 'upload_span')
        @patch.object(openai.Moderation, 'create')
        async def test_moderation_create(self, mocked_create, mocked_upload_span):
            # mocking overrides autoinstrumentation, reinstrument
            recorder = OpenAIRecorder()
            recorder.setup()

            mocked_create.return_value = {
                "id": "modr-id",
                "model": "text-moderation-004",
                "results": [
                    {
                        "categories": {
                            "hate": False,
                            "hate/threatening": False,
                            "self-harm": False,
                            "sexual": False,
                            "sexual/minors": False,
                            "violence": False,
                            "violence/graphic": False
                        },
                        "category_scores": {
                            "hate": 1.5125541722227354e-05,
                            "hate/threatening": 4.6567969036459544e-08,
                            "self-harm": 4.265221065224978e-09,
                            "sexual": 0.00031956686871126294,
                            "sexual/minors": 4.4396051634976175e-06,
                            "violence": 6.315908080978261e-07,
                            "violence/graphic": 6.493833097920287e-07
                        },
                        "flagged": False
                    }
                ]
            }

            response = openai.Moderation.create(
                model='text-moderation-latest',
                input='test text')

            recorder.shutdown()

            proto = mocked_upload_span.call_args[0][0]

            self.assertEqual(find_tag(proto, 'component'), 'Model')
            self.assertEqual(find_tag(proto, 'operation'), 'openai.Moderation.create')
            self.assertEqual(find_tag(proto, 'endpoint'), 'https://api.openai.com/v1/moderations')

            self.assertEqual(find_param(proto, 'model'), 'text-moderation-latest')

            self.assertIsNotNone(find_data_sample(proto, 'input'))



import unittest
import logging
import sys
import os
import json
import time
from unittest.mock import patch, Mock
from google.protobuf.json_format import MessageToJson
import pprint
import openai
from typing import Any, List, Mapping, Optional
from llama_index import GPTVectorStoreIndex, ServiceContext
from llama_index.embeddings.base import BaseEmbedding
from llama_index.readers.schema.base import Document
from llama_index.llm_predictor.structured import LLMPredictor
from langchain.llms.fake import FakeListLLM

import graphsignal
from graphsignal.uploader import Uploader
from graphsignal.callbacks.llama_index.v1 import GraphsignalCallbackHandler
from graphsignal.recorders.openai_recorder import OpenAIRecorder
from graphsignal.proto_utils import find_tag, find_data_count, find_data_sample

logger = logging.getLogger('graphsignal')

class TestEmbedding(BaseEmbedding):
    def _get_text_embedding(self, text: str) -> List[float]:
        embedding_map: Dict[str, List[float]] = {
            "one": [0.5, 0.5],
            "two": [1.0, 1.0]
        }

        return embedding_map[text]

    def _get_query_embedding(self, query: str) -> List[float]:
        embedding_map: Dict[str, List[float]] = {
            "one": [0.5, 0.5],
            "two": [1.0, 1.0]
        }

        return embedding_map[query]

    async def _aget_query_embedding(self, query: str) -> List[float]:
        pass

    async def _aget_text_embedding(self, text: str) -> List[float]:
        pass

class LlamaIndexCallbackHandlerTest(unittest.IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        if len(logger.handlers) == 0:
            logger.addHandler(logging.StreamHandler(sys.stdout))
        graphsignal.configure(
            api_key='k1',
            deployment='d1',
            upload_on_shutdown=False,
            debug_mode=True)

    async def asyncTearDown(self):
        graphsignal.shutdown()

    @patch.object(Uploader, 'upload_span')
    async def test_callback(self, mocked_upload_span):
        os.environ['OPENAI_API_KEY'] = 'sk-kkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkk'

        graphsignal.set_context_tag('ct1', 'v1')

        documents = [Document()]
        documents[0].text = "one"
        service_context = ServiceContext.from_defaults()
        service_context.embed_model = TestEmbedding(callback_manager=service_context.callback_manager)
        llm = FakeListLLM(responses=['three'])
        predictor = LLMPredictor(llm, callback_manager=service_context.callback_manager)
        service_context.llm_predictor = predictor

        index = GPTVectorStoreIndex.from_documents(documents, service_context=service_context)

        query_engine = index.as_query_engine()
        query_engine.query("two")

        #for call in mocked_upload_span.call_args_list:
        #    print(find_tag(call[0][0], 'operation'))

        index_root_span = find_call_by_operation(mocked_upload_span.call_args_list, 'llama_index.index_construction')
        self.assertEqual(index_root_span.labels, ['root'])
        self.assertEqual(find_tag(index_root_span, 'ct1'), 'v1')

        node_parsing_span = find_call_by_operation(mocked_upload_span.call_args_list, 'llama_index.op.node_parsing')
        self.assertEqual(find_tag(node_parsing_span, 'ct1'), 'v1')
        self.assertEqual(find_data_count(node_parsing_span, 'documents', 'element_count'), 1)
        self.assertEqual(node_parsing_span.context.parent_span_id, index_root_span.span_id)
        self.assertEqual(node_parsing_span.context.root_span_id, index_root_span.span_id)

        query_root_span = find_call_by_operation(mocked_upload_span.call_args_list, 'llama_index.query')
        self.assertEqual(query_root_span.labels, ['root'])
        self.assertEqual(find_tag(query_root_span, 'ct1'), 'v1')

        query_span = find_call_by_operation(mocked_upload_span.call_args_list, 'llama_index.op.query')
        self.assertEqual(find_tag(query_span, 'component'), 'Memory')
        self.assertEqual(find_tag(query_span, 'ct1'), 'v1')
        self.assertEqual(query_span.context.parent_span_id, query_root_span.span_id)
        self.assertEqual(query_span.context.root_span_id, query_root_span.span_id)

        retrieve_span = find_call_by_operation(mocked_upload_span.call_args_list, 'llama_index.op.retrieve')
        self.assertEqual(retrieve_span.labels, [])
        self.assertEqual(find_tag(retrieve_span, 'component'), 'Memory')
        self.assertEqual(find_tag(retrieve_span, 'ct1'), 'v1')
        self.assertEqual(retrieve_span.context.parent_span_id, query_span.span_id)
        self.assertEqual(retrieve_span.context.root_span_id, query_root_span.span_id)

        embedding_span = find_call_by_operation(mocked_upload_span.call_args_list, 'llama_index.op.embedding')
        self.assertEqual(embedding_span.labels, [])
        self.assertEqual(find_tag(embedding_span, 'component'), 'LLM')
        self.assertEqual(find_tag(embedding_span, 'ct1'), 'v1')
        self.assertEqual(retrieve_span.context.parent_span_id, query_span.span_id)
        self.assertEqual(retrieve_span.context.root_span_id, query_root_span.span_id)

        synthesize_span = find_call_by_operation(mocked_upload_span.call_args_list, 'llama_index.op.synthesize')
        self.assertEqual(synthesize_span.labels, [])
        self.assertEqual(find_tag(synthesize_span, 'component'), 'Memory')
        self.assertEqual(find_tag(synthesize_span, 'ct1'), 'v1')
        self.assertEqual(find_data_count(synthesize_span, 'response', 'char_count'), 5)
        self.assertEqual(synthesize_span.context.parent_span_id, query_span.span_id)
        self.assertEqual(synthesize_span.context.root_span_id, query_root_span.span_id)

        llm_span = find_call_by_operation(mocked_upload_span.call_args_list, 'llama_index.op.llm')
        self.assertEqual(llm_span.labels, [])
        self.assertEqual(find_tag(llm_span, 'component'), 'LLM')
        self.assertEqual(find_tag(llm_span, 'ct1'), 'v1')
        #self.assertEqual(find_data_count(llm_span, 'formatted_prompt', 'char_count'), 170)
        self.assertIsNotNone(find_data_sample(llm_span, 'formatted_prompt'))
        self.assertIsNotNone(find_data_sample(llm_span, 'completion'))
        self.assertEqual(llm_span.context.parent_span_id, synthesize_span.span_id)
        self.assertEqual(llm_span.context.root_span_id, query_root_span.span_id)

        fake_llm_span = find_call_by_operation(mocked_upload_span.call_args_list, 'langchain_community.llms.fake.FakeListLLM')
        self.assertEqual(fake_llm_span.labels, [])
        self.assertEqual(find_tag(fake_llm_span, 'ct1'), 'v1')
        self.assertEqual(fake_llm_span.context.parent_span_id, llm_span.span_id)
        self.assertEqual(fake_llm_span.context.root_span_id, query_root_span.span_id)


def find_call_by_operation(calls, operation):
    for call in calls:
        if find_tag(call[0][0], 'operation') == operation:
            return call[0][0]


import logging
import llama_index

import graphsignal
from graphsignal.recorders.base_recorder import BaseRecorder
from graphsignal.proto_utils import parse_semver
from graphsignal.proto import signals_pb2
from graphsignal.recorders.instrumentation import patch_method

logger = logging.getLogger('graphsignal')


class LlamaIndexRecorder(BaseRecorder):
    def __init__(self):
        self._library = None
        self._v1_handler = None

    def setup(self):
        if not graphsignal._tracer.auto_instrument:
            return

        self._library = signals_pb2.LibraryInfo()
        self._library.name = 'LlamaIndex'
        version = ''
        if hasattr(llama_index, '__version__') and llama_index.__version__:
            parse_semver(self._library.version, llama_index.__version__)
            version = llama_index.__version__

        def is_v1():
            return (
                hasattr(llama_index.indices.service_context, 'ServiceContext')
            )

        if is_v1():
            from graphsignal.callbacks.llama_index.v1 import GraphsignalCallbackHandler
            from llama_index.indices.service_context import ServiceContext
            def after_from_defaults(args, kwargs, ret, exc, context):
                if isinstance(ret, ServiceContext):
                    if not any(isinstance(handler, GraphsignalCallbackHandler) for handler in ret.callback_manager.handlers):
                        ret.callback_manager.add_handler(GraphsignalCallbackHandler())
                else:
                    logger.error(f'Cannot add callback for LlamaIndex {version}')
            if not patch_method(ServiceContext, 'from_defaults', after_func=after_from_defaults):
                logger.error(f'Cannot instrument LlamaIndex {version}')
        else:
            logger.error(f'Cannot auto-instrument LlamaIndex {version}')

    def shutdown(self):
        if self._v1_handler:
            llama_index.callbacks.get_callback_manager().remove_handler(self._v1_handler)
            self._v1_handler = None

    def on_span_read(self, proto, context, options):
        if self._library:
            proto.libraries.append(self._library)



import logging
from typing import Any, Dict, List, Optional, Union
from uuid import UUID
import contextvars
from llama_index.callbacks.base import BaseCallbackHandler
from llama_index.callbacks.schema import CBEventType

import graphsignal
from graphsignal.recorders.base_recorder import BaseRecorder
from graphsignal.spans import get_current_span, push_current_span, clear_span_stack

logger = logging.getLogger('graphsignal')

# prevent memory leak, if spans are not stopped for some reason
MAX_TRACES = 10000 

# is thread-safe, because event_id is unique per thread
_span_map = {}

_llama_trace_map = contextvars.ContextVar('llama_trace_map', default={})


class GraphsignalCallbackHandler(BaseCallbackHandler):
    def __init__(self, **kwargs: Any) -> None:
        super().__init__(
            event_starts_to_ignore=[],
            event_ends_to_ignore=[]
        )

    def _start_trace(self, event_id, operation):
        if event_id in _span_map or len(_span_map) > MAX_TRACES:
            return None

        span = graphsignal.start_trace(operation)
        _span_map[event_id] = span
        return span

    def _current_span(self, event_id):
        return _span_map.get(event_id)

    def _stop_trace(self, event_id):
        span = _span_map.pop(event_id, None)
        if span:
            span.stop()

    def on_event_start(
        self,
        event_type: CBEventType,
        payload: Optional[Dict[str, Any]] = None,
        event_id: str = "",
        **kwargs: Any
    ) -> str:
        try:
            event_name_str = event_type.name.lower()
            operation = f'llama_index.op.{event_name_str}'

            span = self._start_trace(event_id, operation)
            if span:
                if event_type == CBEventType.CHUNKING:
                    if payload and 'text' in payload:
                        span.set_data('text', payload['text'])
                elif event_type == CBEventType.NODE_PARSING:
                    if payload and 'documents' in payload:
                        span.set_data('documents', payload['documents'])
                elif event_type == CBEventType.EMBEDDING:
                    span.set_tag('component', 'LLM')
                elif event_type == CBEventType.LLM:
                    span.set_tag('component', 'LLM')
                    if payload and 'template' in payload:
                        span.set_data('template', payload['template'])
                    if payload and 'context_str' in payload:
                        span.set_data('context', payload['context_str'])
                elif event_type == CBEventType.QUERY:
                    span.set_tag('component', 'Memory')
                elif event_type == CBEventType.RETRIEVE:
                    span.set_tag('component', 'Memory')
                elif event_type == CBEventType.SYNTHESIZE:
                    span.set_tag('component', 'Memory')
                elif event_type == CBEventType.TREE:
                    pass
        except Exception:
            logger.error('Error in LlamaIndex callback handler', exc_info=True)

    def on_event_end(
        self,
        event_type: CBEventType,
        payload: Optional[Dict[str, Any]] = None,
        event_id: str = "",
        **kwargs: Any
    ) -> None:
        try:
            span = self._current_span(event_id)
            if span:
                if event_type == CBEventType.CHUNKING:
                    if payload and 'chunks' in payload:
                        span.set_data('chunks', payload['chunks'])
                elif event_type == CBEventType.NODE_PARSING:
                    if payload and 'nodes' in payload:
                        span.set_data('nodes',payload['nodes'])
                elif event_type == CBEventType.EMBEDDING:
                    pass
                elif event_type == CBEventType.LLM:
                    if payload:
                        if 'formatted_prompt' in payload:
                            span.set_data('formatted_prompt', payload['formatted_prompt'])
                        if 'response' in payload:
                            span.set_data('response', payload['response'])
                        if 'completion' in payload:
                            span.set_data('completion', payload['completion'])
                elif event_type == CBEventType.QUERY:
                    pass
                elif event_type == CBEventType.RETRIEVE:
                    if payload and 'nodes' in payload:
                        span.set_data('nodes',payload['nodes'])
                elif event_type == CBEventType.SYNTHESIZE:
                    if payload and 'response' in payload:
                        span.set_data('response', str(payload['response']))
                elif event_type == CBEventType.TREE:
                    if payload and 'chunks' in payload:
                        span.set_data('chunks', payload['chunks'])
            self._stop_trace(event_id)
        except Exception:
            logger.error('Error in LlamaIndex callback handler', exc_info=True)

    def start_trace(self, trace_id: Optional[str] = None) -> None:
        try:
            if trace_id is not None:
                trace_map = _llama_trace_map.get()
                if trace_id not in trace_map:
                    operation = f'llama_index.{trace_id}' if trace_id else 'llama_index.root'
                    span = graphsignal.start_trace(operation)
                    trace_map[trace_id] = span
                    _llama_trace_map.set(trace_map)
        except Exception:
            logger.error('Error in LlamaIndex callback handler', exc_info=True)

    def end_trace(
        self,
        trace_id: Optional[str] = None,
        trace_map: Optional[Dict[str, List[str]]] = None,
    ) -> None:
        try:
            if trace_id is not None:
                trace_map = _llama_trace_map.get()
                span = trace_map.pop(trace_id)
                if span is not None:
                    span.stop()
                    _llama_trace_map.set(trace_map)
        except Exception:
            logger.error('Error in LlamaIndex callback handler', exc_info=True)



import logging
import time

logger = logging.getLogger('graphsignal')

class RandomSampler:
    MAX_EVENTS_PER_MIN = 1000

    def __init__(self):
        self._num_events = 0
        self._last_reset_ts = time.time()

    def sample(self):
        now = time.time()
        if now - self._last_reset_ts > 60:
            self._num_events = 0
            self._last_reset_ts = now

        if self._num_events >= self.MAX_EVENTS_PER_MIN:
            return False

        self._num_events += 1
        return True


import logging
import sys
import time

import graphsignal
from graphsignal.spans import TraceOptions
from graphsignal.recorders.base_recorder import BaseRecorder
from graphsignal.recorders.instrumentation import instrument_method, uninstrument_method, read_args
from graphsignal.proto_utils import parse_semver, compare_semver
from graphsignal.proto import signals_pb2
from graphsignal.proto_utils import add_library_param, add_driver

logger = logging.getLogger('graphsignal')

class BananaRecorder(BaseRecorder):
    def __init__(self):
        self._library = None

    def setup(self):
        if not graphsignal._tracer.auto_instrument:
            return

        self._library = signals_pb2.LibraryInfo()
        self._library.name = 'Banana Python SDK'

        from banana_dev import Client
        instrument_method(Client, 'call', 'banana.call', trace_func=self.trace_call)

    def shutdown(self):
        from banana_dev import Client
        uninstrument_method(Client, 'call')

    def trace_call(self, span, args, kwargs, ret, exc):
        client = args[0]
        params = read_args(args, kwargs, ['self', 'route', 'json'])

        url = client.url if client.url else ''
        endpoint = url + params.get('route', '')

        span.set_tag('component', 'Model')
        if endpoint:
            span.set_tag('endpoint', endpoint)

        if client.model_key:
            span.set_tag('model_key', client.model_key)
            span.set_param('model_key', client.model_key)

        if 'json' in params:
            span.set_data('json', params['json'])
        if 'headers' in params:
            span.set_data('headers', params['headers'])

        if isinstance(ret, tuple):
            span.set_data('output', ret[0])
            if len(ret) > 1:
                span.set_data('meta', ret[1])

    def on_span_read(self, proto, context, options):
        if self._library:
            proto.libraries.append(self._library)



import unittest
import logging
import sys
import os
import json
import time
from unittest.mock import patch, Mock
from google.protobuf.json_format import MessageToJson
import pprint
import types
import openai

import graphsignal
from graphsignal.proto import signals_pb2
from graphsignal.uploader import Uploader
from graphsignal.spans import DEFAULT_OPTIONS
from graphsignal.recorders.openai_recorder import OpenAIRecorder
from graphsignal.proto_utils import find_tag, find_param, find_data_count, find_data_sample
from graphsignal.proto_utils import parse_semver, compare_semver
from graphsignal.proto import signals_pb2

logger = logging.getLogger('graphsignal')

openai_lib = signals_pb2.LibraryInfo()
parse_semver(openai_lib.version, openai.version.VERSION)

if compare_semver(openai_lib.version, (1, 0, 0)) >= 0:
    from openai import OpenAI, AsyncOpenAI
    from openai.types import Completion, CompletionChoice, CompletionUsage, CreateEmbeddingResponse, Embedding
    from openai.types.chat import ChatCompletion, ChatCompletionMessage, ChatCompletionChunk
    from openai.types.chat.chat_completion import Choice
    from openai.types.chat.chat_completion_chunk import Choice as ChunkChoice, ChoiceDelta
    from openai.types.chat.chat_completion_message import FunctionCall
    from openai.types.create_embedding_response import Usage

    os.environ['OPENAI_API_KEY'] = 'sk-kkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkk'

    class OpenAIRecorderTest(unittest.IsolatedAsyncioTestCase):
        async def asyncSetUp(self):
            if len(logger.handlers) == 0:
                logger.addHandler(logging.StreamHandler(sys.stdout))
            graphsignal.configure(
                api_key='k1',
                deployment='d1',
                upload_on_shutdown=False,
                debug_mode=True)

        async def asyncTearDown(self):
            graphsignal.shutdown()

        @patch.object(Uploader, 'upload_span')
        async def test_record(self, mocked_upload_span):
            recorder = OpenAIRecorder()
            recorder.setup()
            proto = signals_pb2.Span()
            context = {}
            recorder.on_span_start(proto, context, DEFAULT_OPTIONS)
            recorder.on_span_stop(proto, context, DEFAULT_OPTIONS)
            recorder.on_span_read(proto, context, DEFAULT_OPTIONS)
            self.assertEqual(proto.libraries[0].name, 'OpenAI Python Library')

        @patch.object(Uploader, 'upload_span')
        async def test_client(self, mocked_upload_span):
            recorder = OpenAIRecorder()
            recorder.setup()
            proto = signals_pb2.Span()

            context = {}
            recorder.on_span_start(proto, context, DEFAULT_OPTIONS)
            recorder.on_span_stop(proto, context, DEFAULT_OPTIONS)
            recorder.on_span_read(proto, context, DEFAULT_OPTIONS)
            self.assertEqual(proto.libraries[0].name, 'OpenAI Python Library')

        @patch.object(Uploader, 'upload_span')
        async def test_completion_create(self, mocked_upload_span):
            recorder = OpenAIRecorder()
            recorder.setup()

            client = OpenAI(
                #api_key="Your_API_Key"
            )

            with patch.object(client.completions, '_post') as mocked_post:
                mocked_post.return_value = Completion(
                    id='cmpl-8Igp0TaBfwDxyxZlB2Lcgrtg89HmQ', 
                    choices=[
                        CompletionChoice(
                            finish_reason='stop', 
                            index=0, 
                            logprobs=None, 
                            text='\n\nWhat is your favorite color?'), 
                        CompletionChoice(
                            finish_reason='stop', 
                            index=1, logprobs=None, 
                            text='\n\nWhat is your favorite color?\n\nMy favorite color is blue.')], 
                    created=1699465202, 
                    model='gpt-3.5-turbo-instruct', 
                    object='text_completion', 
                    system_fingerprint=None, 
                    usage=CompletionUsage(
                        completion_tokens=20, 
                        prompt_tokens=8, 
                        total_tokens=28))

                response = client.completions.create(
                    model="gpt-3.5-turbo-instruct", 
                    prompt=['test prompt 1', 'test prompt 2'],
                    temperature=0.1,
                    top_p=1,
                    max_tokens=1024,
                    frequency_penalty=0,
                    presence_penalty=0)

                recorder.shutdown()

                #print(response.model_dump())

                proto = mocked_upload_span.call_args[0][0]

                self.assertEqual(find_tag(proto, 'component'), 'LLM')
                self.assertEqual(find_tag(proto, 'operation'), 'openai.completions.create')
                self.assertEqual(find_tag(proto, 'endpoint'), 'https://api.openai.com/v1/completions')
                self.assertEqual(find_tag(proto, 'model'), 'gpt-3.5-turbo-instruct')

                self.assertEqual(find_param(proto, 'model'), 'gpt-3.5-turbo-instruct')
                self.assertEqual(find_param(proto, 'max_tokens'), '1024')
                self.assertEqual(find_param(proto, 'temperature'), '0.1')
                self.assertEqual(find_param(proto, 'top_p'), '1')
                self.assertEqual(find_param(proto, 'presence_penalty'), '0')
                self.assertEqual(find_param(proto, 'frequency_penalty'), '0')

                self.assertEqual(find_data_count(proto, 'prompt', 'token_count'), 8)
                self.assertEqual(find_data_count(proto, 'completion', 'token_count'), 20)
                self.assertEqual(find_data_count(proto, 'completion', 'finish_reason_stop'), 2)

                self.assertIsNotNone(find_data_sample(proto, 'prompt'))
                self.assertIsNotNone(find_data_sample(proto, 'completion'))

        @patch.object(Uploader, 'upload_span')
        async def test_completion_create_stream(self, mocked_upload_span):
            recorder = OpenAIRecorder()
            recorder.setup()

            client = OpenAI(
                #api_key="Your_API_Key"
            )

            with patch.object(client.completions, '_post') as mocked_post:
                test_ret = [
                    Completion(
                        id='cmpl-8NH2zI1lzfUs1SId2EDIFAOkcubN8', 
                        choices=[
                            CompletionChoice(
                                finish_reason='stop', 
                                index=0, 
                                logprobs=None, 
                                text=' to')], 
                                created=1700557765, 
                                model='gpt-3.5-turbo-instruct', 
                                object='text_completion', 
                                system_fingerprint=None, 
                                usage=None),
                    Completion(
                        id='cmpl-8NH2zI1lzfUs1SId2EDIFAOkcubN8', 
                        choices=[
                            CompletionChoice(
                                finish_reason='stop', 
                                index=0, 
                                logprobs=None, 
                                text=' ')], 
                                created=1700557765, 
                                model='gpt-3.5-turbo-instruct', 
                                object='text_completion', 
                                system_fingerprint=None, 
                                usage=None)
                ]
                def test_ret_gen():
                    for item in test_ret:
                        yield item
                mocked_post.return_value = test_ret_gen()

                response = client.completions.create(
                    model="gpt-3.5-turbo-instruct", 
                    prompt=['count 1 to 3', 'generate 2 random letters'],
                    temperature=0.1,
                    top_p=1,
                    max_tokens=1024,
                    frequency_penalty=0,
                    presence_penalty=0,
                    stream=True)

                for r in response:
                    pass

                recorder.shutdown()

                proto = mocked_upload_span.call_args[0][0]

                self.assertEqual(find_data_count(proto, 'prompt', 'token_count'), 11)
                self.assertEqual(find_data_count(proto, 'completion', 'finish_reason_stop'), 2)
                self.assertEqual(find_data_count(proto, 'completion', 'token_count'), 2)

                self.assertIsNotNone(find_data_sample(proto, 'prompt'))
                self.assertIsNotNone(find_data_sample(proto, 'completion'))

        @patch.object(Uploader, 'upload_span')
        async def test_completion_create_async(self, mocked_upload_span):
            recorder = OpenAIRecorder()
            recorder.setup()

            client = AsyncOpenAI(
                #api_key="Your_API_Key"
            )

            with patch.object(client.completions, '_post') as mocked_post:
                mocked_post.return_value = Completion(
                    id='cmpl-8Igp0TaBfwDxyxZlB2Lcgrtg89HmQ', 
                    choices=[
                        CompletionChoice(
                            finish_reason='stop', 
                            index=0, 
                            logprobs=None, 
                            text='\n\nWhat is your favorite color?'), 
                        CompletionChoice(
                            finish_reason='stop', 
                            index=1, logprobs=None, 
                            text='\n\nWhat is your favorite color?\n\nMy favorite color is blue.')], 
                    created=1699465202, 
                    model='gpt-3.5-turbo-instruct', 
                    object='text_completion', 
                    system_fingerprint=None, 
                    usage=CompletionUsage(
                        completion_tokens=20, 
                        prompt_tokens=8, 
                        total_tokens=28))

                response = await client.completions.create(
                    model="gpt-3.5-turbo-instruct", 
                    prompt=['test prompt 1', 'test prompt 2'],
                    temperature=0.1,
                    top_p=1,
                    max_tokens=1024,
                    frequency_penalty=0,
                    presence_penalty=0)

                recorder.shutdown()

                #print(response.model_dump())

                proto = mocked_upload_span.call_args[0][0]

                self.assertEqual(find_tag(proto, 'component'), 'LLM')
                self.assertEqual(find_tag(proto, 'operation'), 'openai.completions.create')
                self.assertEqual(find_tag(proto, 'endpoint'), 'https://api.openai.com/v1/completions')
                self.assertEqual(find_tag(proto, 'model'), 'gpt-3.5-turbo-instruct')

                self.assertEqual(find_param(proto, 'model'), 'gpt-3.5-turbo-instruct')
                self.assertEqual(find_param(proto, 'max_tokens'), '1024')
                self.assertEqual(find_param(proto, 'temperature'), '0.1')
                self.assertEqual(find_param(proto, 'top_p'), '1')
                self.assertEqual(find_param(proto, 'presence_penalty'), '0')
                self.assertEqual(find_param(proto, 'frequency_penalty'), '0')

                self.assertEqual(find_data_count(proto, 'prompt', 'token_count'), 8)
                self.assertEqual(find_data_count(proto, 'completion', 'token_count'), 20)
                self.assertEqual(find_data_count(proto, 'completion', 'finish_reason_stop'), 2)

                self.assertIsNotNone(find_data_sample(proto, 'prompt'))
                self.assertIsNotNone(find_data_sample(proto, 'completion'))


        @patch.object(Uploader, 'upload_span')
        async def test_chat_completion_create(self, mocked_upload_span):
            recorder = OpenAIRecorder()
            recorder.setup()

            client = OpenAI(
                #api_key="Your_API_Key"
            )

            with patch.object(client.chat.completions, '_post') as mocked_post:
                mocked_post.return_value = ChatCompletion(
                    id='chatcmpl-8NeQhqaA8Cq3fVeJpznStwgC4hWxt', 
                    choices=[
                        Choice(
                            finish_reason='stop', 
                            index=0, 
                            message=ChatCompletionMessage(
                                content='What are the benefits of regular exercise?', 
                                role='assistant', 
                                function_call=None, 
                                tool_calls=None))], 
                                created=1700647647, 
                                model='gpt-3.5-turbo-0613', 
                                object='chat.completion', 
                                system_fingerprint=None, 
                                usage=CompletionUsage(
                                    completion_tokens=8, 
                                    prompt_tokens=19, 
                                    total_tokens=27))

                response = client.chat.completions.create(
                    model="gpt-3.5-turbo", 
                    messages=[
                        {'role': 'system', 'content': 'test prompt 1'}, 
                        {'role': 'system', 'content': 'test prompt 2'}],
                    temperature=0.1,
                    top_p=1,
                    max_tokens=1024,
                    frequency_penalty=0,
                    presence_penalty=0)

                recorder.shutdown()

                proto = mocked_upload_span.call_args[0][0]

                self.assertEqual(find_tag(proto, 'component'), 'LLM')
                self.assertEqual(find_tag(proto, 'operation'), 'openai.chat.completions.create')
                self.assertEqual(find_tag(proto, 'endpoint'), 'https://api.openai.com/v1/chat/completions')
                self.assertEqual(find_tag(proto, 'model'), 'gpt-3.5-turbo')

                self.assertEqual(find_param(proto, 'model'), 'gpt-3.5-turbo')
                self.assertEqual(find_param(proto, 'max_tokens'), '1024')
                self.assertEqual(find_param(proto, 'temperature'), '0.1')
                self.assertEqual(find_param(proto, 'top_p'), '1')
                self.assertEqual(find_param(proto, 'presence_penalty'), '0')
                self.assertEqual(find_param(proto, 'frequency_penalty'), '0')

                self.assertEqual(find_data_count(proto, 'messages', 'token_count'), 19)
                self.assertEqual(find_data_count(proto, 'completion', 'token_count'), 8)
                self.assertEqual(find_data_count(proto, 'completion', 'finish_reason_stop'), 1.0)

                self.assertIsNotNone(find_data_sample(proto, 'messages'))
                self.assertIsNotNone(find_data_sample(proto, 'completion'))


        @patch.object(Uploader, 'upload_span')
        async def test_chat_completion_create_function(self, mocked_upload_span):
            recorder = OpenAIRecorder()
            recorder.setup()

            client = OpenAI(
                #api_key="Your_API_Key"
            )

            with patch.object(client.chat.completions, '_post') as mocked_post:
                mocked_post.return_value = ChatCompletion(
                    id='chatcmpl-8NjG5hzrQXtbnnYFsR3t3siBRcFWl', 
                    choices=[
                        Choice(
                            finish_reason='function_call', 
                            index=0, 
                            message=ChatCompletionMessage(
                                content=None, 
                                role='assistant', 
                                function_call=FunctionCall(
                                    arguments='{\n  "location": "Boston, MA"\n}', 
                                    name='get_current_weather'), tool_calls=None))], 
                                    created=1700666209, 
                                    model='gpt-3.5-turbo-0613', 
                                    object='chat.completion', 
                                    system_fingerprint=None, 
                                    usage=CompletionUsage(
                                        completion_tokens=18, 
                                        prompt_tokens=78, 
                                        total_tokens=96))

                response = client.chat.completions.create(
                    model="gpt-3.5-turbo-0613", 
                    messages=[{'role': 'system', 'content': 'What\'s the weather like in Boston?'}],
                    functions=[
                        {
                            "name": "get_current_weather",
                            "description": "Get the current weather in a given location",
                            "parameters": {
                                "type": "object",
                                "properties": {
                                    "location": {
                                        "type": "string",
                                        "description": "The city and state, e.g. San Francisco, CA",
                                    },
                                    "unit": {"type": "string", "enum": ["celsius", "fahrenheit"]},
                                },
                                "required": ["location"],
                            },
                        }
                    ],
                    function_call="auto",
                    temperature=0.1,
                    top_p=1,
                    max_tokens=1024,
                    frequency_penalty=0,
                    presence_penalty=0)

                recorder.shutdown()

                #print(response)

                proto = mocked_upload_span.call_args[0][0]

                self.assertEqual(find_tag(proto, 'component'), 'LLM')
                self.assertEqual(find_tag(proto, 'operation'), 'openai.chat.completions.create')
                self.assertEqual(find_tag(proto, 'endpoint'), 'https://api.openai.com/v1/chat/completions')
                self.assertEqual(find_tag(proto, 'model'), 'gpt-3.5-turbo-0613')

                self.assertEqual(find_param(proto, 'model'), 'gpt-3.5-turbo-0613')
                self.assertEqual(find_param(proto, 'function_call'), 'auto')
                self.assertEqual(find_param(proto, 'max_tokens'), '1024')
                self.assertEqual(find_param(proto, 'temperature'), '0.1')
                self.assertEqual(find_param(proto, 'top_p'), '1')
                self.assertEqual(find_param(proto, 'presence_penalty'), '0')
                self.assertEqual(find_param(proto, 'frequency_penalty'), '0')

                self.assertEqual(find_data_count(proto, 'messages', 'token_count'), 78)
                self.assertEqual(find_data_count(proto, 'completion', 'token_count'), 18)
                self.assertEqual(find_data_count(proto, 'completion', 'finish_reason_function_call'), 1.0)

                self.assertIsNotNone(find_data_sample(proto, 'messages'))
                self.assertIsNotNone(find_data_sample(proto, 'functions'))
                self.assertIsNotNone(find_data_sample(proto, 'completion'))


        @patch.object(Uploader, 'upload_span')
        async def test_chat_completion_create_stream(self, mocked_upload_span):
            recorder = OpenAIRecorder()
            recorder.setup()

            client = OpenAI(
                #api_key="Your_API_Key"
            )

            with patch.object(client.chat.completions, '_post') as mocked_post:
                test_ret = [
                    ChatCompletionChunk(
                        id='chatcmpl-8O3j3TnSiuvjZTBeBFx3aWA9Ukh2x',
                        choices=[
                            ChunkChoice(
                                delta=ChoiceDelta(
                                    content='', 
                                    function_call=None, 
                                    role='assistant', 
                                    tool_calls=None), 
                                finish_reason='stop', 
                                index=0)], 
                            created=1700744885, 
                            model='gpt-4-0613', 
                            object='chat.completion.chunk'),
                    ChatCompletionChunk(
                        id='chatcmpl-8O3j3TnSiuvjZTBeBFx3aWA9Ukh2x', 
                        choices=[
                            ChunkChoice(
                                delta=ChoiceDelta(
                                    content='We', 
                                    function_call=None, 
                                    role=None, 
                                    tool_calls=None), 
                                finish_reason='stop', 
                                index=0)], 
                            created=1700744885, 
                            model='gpt-4-0613', 
                            object='chat.completion.chunk')
                ]
                def test_ret_gen():
                    for item in test_ret:
                        yield item
                mocked_post.return_value = test_ret_gen()

                response = client.chat.completions.create(
                    model="gpt-4", 
                    messages=[
                        {
                            "role": "system",
                            "content": "You are a helpful.",
                        },
                        {
                            "role": "system",
                            "name": "example_user",
                            "content": "New synergies will help drive top-line growth.",
                        },
                        {
                            "role": "user",
                            "name": "example",
                            "content": "This late pivot means.",
                        }
                    ],
                    temperature=0.1,
                    top_p=1,
                    max_tokens=1024,
                    frequency_penalty=0,
                    presence_penalty=0,
                    stream=True)

                for r in response:
                    pass

                recorder.shutdown()

                proto = mocked_upload_span.call_args[0][0]

                self.assertEqual(find_data_count(proto, 'messages', 'token_count'), 40)
                self.assertEqual(find_data_count(proto, 'completion', 'finish_reason_stop'), 2)
                self.assertEqual(find_data_count(proto, 'completion', 'token_count'), 2)

                self.assertIsNotNone(find_data_sample(proto, 'messages'))
                self.assertIsNotNone(find_data_sample(proto, 'completion'))


        @patch.object(Uploader, 'upload_span')
        async def test_chat_completion_create_stream_async(self, mocked_upload_span):
            recorder = OpenAIRecorder()
            recorder.setup()

            client = AsyncOpenAI(
                #api_key="Your_API_Key"
            )

            with patch.object(client.chat.completions, '_post') as mocked_post:
                test_ret = [
                    ChatCompletionChunk(
                        id='chatcmpl-8OQctcUZdbOT1KO8x2IKGUQE7G33b', 
                        choices=[ChunkChoice(
                            delta=ChoiceDelta(
                                content='', 
                                function_call=None, 
                                role='assistant', 
                                tool_calls=None), 
                            finish_reason='stop', 
                            index=0)], 
                        created=1700832915, 
                        model='gpt-4-0613', 
                        object='chat.completion.chunk'),
                    ChatCompletionChunk(
                        id='chatcmpl-8OQctcUZdbOT1KO8x2IKGUQE7G33b', 
                        choices=[ChunkChoice(
                            delta=ChoiceDelta(
                                content='We', 
                                function_call=None, 
                                role=None, 
                                tool_calls=None), 
                            finish_reason='stop', index=0)], 
                        created=1700832915, 
                        model='gpt-4-0613', 
                        object='chat.completion.chunk')                    
                ]

                async def test_ret_gen():
                    for item in test_ret:
                        yield item
                mocked_post.return_value = test_ret_gen()

                response = await client.chat.completions.create(
                    model="gpt-4", 
                    messages=[
                        {
                            "role": "system",
                            "content": "You are a helpful.",
                        },
                        {
                            "role": "system",
                            "name": "example_user",
                            "content": "New synergies will help drive top-line growth.",
                        },
                        {
                            "role": "user",
                            "name": "example",
                            "content": "This late pivot means.",
                        }
                    ],
                    temperature=0.1,
                    top_p=1,
                    max_tokens=1024,
                    frequency_penalty=0,
                    presence_penalty=0,
                    stream=True)

                async for r in response:
                    pass

                recorder.shutdown()

                proto = mocked_upload_span.call_args[0][0]

                self.assertEqual(find_data_count(proto, 'messages', 'token_count'), 40)
                self.assertEqual(find_data_count(proto, 'completion', 'finish_reason_stop'), 2)
                self.assertEqual(find_data_count(proto, 'completion', 'token_count'), 2)

                self.assertIsNotNone(find_data_sample(proto, 'messages'))
                self.assertIsNotNone(find_data_sample(proto, 'completion'))

        @patch.object(Uploader, 'upload_span')
        async def test_embedding_create(self, mocked_upload_span):
            recorder = OpenAIRecorder()
            recorder.setup()

            client = OpenAI(
                #api_key="Your_API_Key"
            )

            with patch.object(client.embeddings, '_post') as mocked_post:
                mocked_post.return_value = CreateEmbeddingResponse(
                    data=[Embedding(
                        embedding=[0.0008772446890361607, -0.0045392164029181], 
                        index=1, 
                        object='embedding')], 
                    model='text-embedding-ada-002-v2', 
                    object='list', 
                    usage=Usage(
                        prompt_tokens=8, 
                        total_tokens=8))

                response = client.embeddings.create(
                    model="text-embedding-ada-002",
                    input=['test text 1', 'test text 2'])

                recorder.shutdown()

                proto = mocked_upload_span.call_args[0][0]

                self.assertEqual(find_tag(proto, 'component'), 'LLM')
                self.assertEqual(find_tag(proto, 'operation'), 'openai.embeddings.create')
                self.assertEqual(find_tag(proto, 'endpoint'), 'https://api.openai.com/v1/embeddings')
                self.assertEqual(find_tag(proto, 'model'), 'text-embedding-ada-002-v2')

                self.assertEqual(find_param(proto, 'model'), 'text-embedding-ada-002')

                self.assertEqual(find_data_count(proto, 'input', 'token_count'), 8.0)

                self.assertIsNotNone(find_data_sample(proto, 'input'))
                self.assertIsNotNone(find_data_sample(proto, 'embeddings'))



import logging
import os
import sys
import platform
import socket
import re

from graphsignal.proto import signals_pb2

logger = logging.getLogger('graphsignal')

version_regexp = re.compile(r'^(\d+)\.?(\d+)?\.?(\d+)?')


def parse_semver(semver_proto, version):
    version_match = version_regexp.match(str(version))
    if version_match is not None:
        groups = version_match.groups()
        if groups[0] is not None:
            semver_proto.major = int(groups[0])
        if groups[1] is not None:
            semver_proto.minor = int(groups[1])
        if groups[2] is not None:
            semver_proto.patch = int(groups[2])


def compare_semver(semver_proto, version):
    semver_int = semver_proto.major * 1e6 + semver_proto.minor * 1e3 + semver_proto.patch
    version_int = version[0] * 1e6 + version[1] * 1e3 + version[2]
    if semver_int < version_int:
        return -1
    if semver_int > version_int:
        return 1
    else:
        return 0


def add_library_param(library_info, name, value):
    param = library_info.params.add()
    param.name = name
    param.value = str(value)


def add_driver(node_info, name, version):
    driver = node_info.drivers.add()
    driver.name = name
    driver.version = version


def find_tag(proto, key):
    for tag in proto.tags:
        if tag.key == key:
            return tag.value
    return None


def find_param(proto, name):
    for param in proto.params:
        if param.name == name:
            return param.value
    return None


def find_data_count(proto, data_name, count_name):
    for data_stats in proto.data_profile:
        if data_stats.data_name == data_name:
            for data_count in data_stats.counts:
                if data_count.name == count_name:
                    return data_count.count
    return None


def find_data_sample(proto, data_name):
    for data_sample in proto.data_samples:
        if data_sample.data_name == data_name:
            return data_sample
    return None


import logging
import transformers

import graphsignal
from graphsignal.spans import TraceOptions
from graphsignal.recorders.base_recorder import BaseRecorder
from graphsignal.recorders.instrumentation import instrument_method, uninstrument_method, read_args
from graphsignal.proto_utils import parse_semver, compare_semver
from graphsignal.proto import signals_pb2

logger = logging.getLogger('graphsignal')

class HuggingFaceRecorder(BaseRecorder):
    def __init__(self):
        self._library = None

    def setup(self):
        if not graphsignal._tracer.auto_instrument:
            return

        self._library = signals_pb2.LibraryInfo()
        self._library.name = 'Transformers'
        parse_semver(self._library.version, transformers.__version__)

        from transformers.tools import Agent, RemoteTool, PipelineTool
        instrument_method(Agent, 'run', 'transformers.tools.Agent.run', trace_func=self.trace_run)
        instrument_method(Agent, 'chat', 'transformers.tools.Agent.chat', trace_func=self.trace_run)

        instrument_method(RemoteTool, '__call__', op_func=self.tool_name, trace_func=self.trace_tool)
        instrument_method(PipelineTool, '__call__', op_func=self.tool_name, trace_func=self.trace_tool)

    def shutdown(self):
        from transformers.tools import Agent, RemoteTool, PipelineTool
        uninstrument_method(Agent, 'run')
        uninstrument_method(Agent, 'chat')
        uninstrument_method(RemoteTool, '__call__')
        uninstrument_method(PipelineTool, '__call__')

    def tool_name(self, args, kwargs):
        class_name = args[0].__class__.__name__
        if hasattr(args[0], 'tool_class') and args[0].tool_class:
            class_name = args[0].tool_class.__name__
        else:
            class_name = args[0].__class__.__name__
        return f'transformers.tools.{class_name}'

    def trace_run(self, span, args, kwargs, ret, exc):
        params = read_args(args, kwargs, ['agent', 'task', 'return_code', 'remote'])

        span.set_tag('component', 'Agent')

        if 'return_code' in params:
            span.set_param('return_code', params['return_code'])

        if 'remote' in params:
            span.set_param('remote', params['remote'])

        if 'task' in params:
            span.set_data('task', params['task'])

        if ret:
            span.set_data('output', ret)

    def trace_tool(self, span, args, kwargs, ret, exc):
        span.set_tag('component', 'Tool')
        if hasattr(args[0], 'endpoint_url'):
            span.set_tag('endpoint', args[0].endpoint_url)

        inputs = {}
        if len(args) > 1:
            inputs['args'] = args[1:]
        if len(kwargs) > 0:
            inputs.update(kwargs)
        if len(inputs) > 0:
            span.set_data('inputs', inputs)

        if ret:
            span.set_data('outputs', ret)

    def on_span_read(self, proto, context, options):
        if self._library:
            proto.libraries.append(self._library)
